import librosa
import numpy as np
import pandas as pd
import soundfile as sf
import pyrubberband as pyrb
from pydub import AudioSegment
from sklearn.cluster import KMeans
from IPython.display import Audio
import scipy
from essentia.standard import *
import essentia.streaming as ess
import essentia.standard as es
from sklearn.preprocessing import StandardScaler
import sounddevice as sd
from typing import Optional, List
from Mix50.process import Process
from Mix50.features import Features

class Effects:

    def __init__(self):
        pass


    def set_audio(self, path, y1, sr1, y2=None, sr2=None):
        '''
        Set the audio from our mixfifty module
        '''
        self.y = y1
        self.sr = sr1
        self.path = path
        

    def __fade_between(self, sample, fade_in_duration, fade_out_duration):
        
        '''
        A helper function to fade between small samples of audio;
        Clicks and pops often occur when overlaying pieces of audio.
        This will prevent that by applying linesplace fade between overlays. 
        '''
        sample_length = len(sample)
        time = np.linspace(0, 1, sample_length)

        fade_in_env = np.linspace(0, 1, int(sample_length * fade_in_duration))
        fade_out_env = np.linspace(1, 0, int(sample_length * fade_out_duration))

        sample[:len(fade_in_env)] *= fade_in_env
        sample[-len(fade_out_env):] *= fade_out_env

        return sample


    def fade_out(self, start_time:int, fade_duration:int) -> object:
        '''
        Apply a fade-out effect to the audio signal.

        Parameters
        ----------
        start_time : int
            The time in seconds from which the fade-out effect should begin.
        fade_duration : int
            The duration in seconds over which the fade-out effect should be applied.

        Notes
        -----
        The fade-out effect gradually reduces the volume of the audio signal to silence
        over the specified duration. The effect is applied starting from the given start time.
        The audio signal is modified in place.

        Examples
        --------
        To fade out an audio signal starting at 10 seconds and lasting for 5 seconds:

        >>> audio.fade_out(start_time=10, fade_duration=5, play=True)

        This will apply a fade-out effect to the audio, beginning at 10 seconds and completing
        the fade-out over a 5-second period.
        '''

        # Get duration and samples per second
        dur = librosa.get_duration(y=self.y, sr=self.sr)
        total_samples = len(self.y)

        # Convert start_time and fade_duration from seconds to sample frames
        start_sample = int(start_time * self.sr)
        fade_samples = int(fade_duration * self.sr)
        end_sample = min(start_sample + fade_samples, total_samples)

        # Create fade-out envelope
        fading_audio = self.y[start_sample:end_sample]

        #Error handling for OOB time arguments
        if len(fading_audio) == 0:
            raise ValueError("The start time is out of bounds; please choose a start time within the length of the audio")

        fade_env = np.linspace(1, 0, num=len(fading_audio))
        fading_audio *= fade_env
        
        #Create silence for the rest of the audio
        silence_audio = np.zeros(total_samples - end_sample)

        # Create full audio with fade-out applied
        full_audio = np.concatenate([
            self.y[:start_sample],
            fading_audio,
            silence_audio
        ])

        return Process(full_audio)
    

    def fade_in(self, start_time:int, fade_duration:int) -> object:
        
        '''
        Apply a fade-in effect to the audio signal.

        Parameters
        ----------
        start_time : int
            The time in seconds from which the fade-in effect should begin.
        fade_duration : int
            The duration in seconds over which the fade-in effect should be applied.

        Notes
        -----
        The fade-in effect gradually increases the volume of the audio signal from silence
        to its original level over the specified duration. The effect is applied starting 
        from the given start time. The audio signal is modified in place.

        Examples
        --------
        To fade in an audio signal starting at 10 seconds and lasting for 5 seconds:

        >>> audio.fade_in(start_time=10, fade_duration=5)

        This will apply a fade-in effect to the audio, beginning at 10 seconds and completing
        the fade-in over a 5-second period.
        '''
        # Get duration and samples per second
        dur = librosa.get_duration(y=self.y, sr=self.sr)
        total_samples = len(self.y)
    
        # Convert start_time and fade_duration from seconds to sample frames
        start_sample = int(start_time * self.sr)
        fade_samples = int(fade_duration * self.sr)
        end_sample = min(start_sample + fade_samples, total_samples)
        
        # Create fade-out envelope
        fading_audio = self.y[start_sample:end_sample]

        #Error handling for OOB time arguments
        if len(fading_audio) == 0:
            raise ValueError("The start time is out of bounds; please choose a start time within the length of the audio")

        
        fade_env = np.linspace(0,1, num=len(fading_audio))
        fading_audio *= fade_env
        
        #Create silence for the start of the audio
        silence_audio = np.zeros(start_sample)

        # Create full audio with fade-out applied
        full_audio = np.concatenate([
            silence_audio,
            fading_audio,
            self.y[end_sample:]
            
        ])

        return Process(full_audio)


    def speed_control(self, start_time:int, end_time:int, original_bpm:int, new_bpm:int) -> object:
        """
        Adjust the speed of a given audio sample between specified start and end times to match a new BPM.

        Parameters
        ----------
        start_time : float
            The start time in seconds of the segment to adjust.
        end_time : float
            The end time in seconds of the segment to adjust.
        original_bpm : float
            The original beats per minute of the audio sample.
        new_bpm : float
            The target beats per minute for the audio sample.

        Notes
        -----
        This function modifies the playback speed of an audio sample by interpolating between different speed rates
        to smoothly transition from the original BPM to the new BPM over the specified time range. The result is a
        time-stretched and faded audio segment that blends with the original audio before and after the adjusted segment.

        Examples
        --------
        To adjust the speed of an audio sample from 120 BPM to 150 BPM, applying the effect between 30 and 60 seconds:

        >>> audio.speed_control(start_time=30, end_time=60, original_bpm=120, new_bpm=150)

        This will modify the playback speed of the audio segment between 30 and 60 seconds to transition smoothly from
        120 BPM to 150 BPM. If `play` is set to True, the modified audio will be played; otherwise, the raw modified audio
        will be returned.
        """
        num_fragments = 15 #Too high fragments will create clicks/pops

        #Calculate Speed Rate and chunk into speed fram
        speed_rate = (new_bpm/original_bpm)
        speed_frags = abs((1-speed_rate) / num_fragments)
        
        time_rates = [] #Get time rates
        start_rate = 1 #Start at current bpm

        #Append Speed rates over 15 intervals
        for i in range(num_fragments):
            start_rate += speed_frags if speed_rate >= 1 else -speed_frags
            time_rates.append(start_rate)

            
        y_new = [] #Create list to store stretched audio
        speed_sample = self.y[int(start_time*self.sr): int(end_time*self.sr)]
        frame_size = len(speed_sample) / num_fragments

        before_sample = self.y[:int(start_time*self.sr)]
        after_sample = pyrb.time_stretch(self.y[int(end_time*self.sr):], sr=self.sr, rate=speed_rate)
        pct = (end_time - start_time )/ num_fragments
        start = 0
        end = num_fragments

        #Append stretched audio to y_new
        for fragment,time_rate in zip(range(num_fragments),time_rates):
            yn = speed_sample[int(fragment*frame_size): int((fragment+1)*frame_size)]
            yn = pyrb.time_stretch(yn, sr=self.sr,rate=time_rate)
            yn = self.__fade_between(yn, .003, .003) #Fade between as to prevent pops/clicks
            y_new += (list(yn))

        #Create final new audio
        y_final = np.array(y_new)
        y_final = np.concatenate((before_sample, y_final, after_sample))

        return Process(y_final)


    def highpass_control(self, start_time:int, end_time:int, cutoff_freq:int, order=5) -> object:

        """
        Applies a progressive high-pass filter to an audio segment between specified start and end times.

        Parameters
        ----------
        start_time : int
            The start time in seconds of the segment to apply the high-pass filter.
        end_time : int
            The end time in seconds of the segment to apply the high-pass filter.
        cutoff_freq : int
            The cutoff frequency of the high-pass filter in Hertz. The filter will progressively adjust this frequency.
        order : int, optional
            The order of the Butterworth filter. Default is 5. Higher values result in a steeper roll-off.

        Returns
        -------
        numpy.ndarray or Audio
            If `play` is True, returns an `Audio` object with the processed audio. If `play` is False, returns the raw
            processed audio data as a numpy array.

        Notes
        -----
        The function applies a high-pass filter to the specified audio segment, starting with a low cutoff frequency
        and progressively increasing it. This is done in 50 intervals, creating a smooth transition effect. The audio
        segments before and after the specified range are not filtered.

        Examples
        --------
        To apply a high-pass filter starting at 10 seconds, ending at 20 seconds, with a cutoff frequency of 500 Hz:

        >>> audio.highpass_control(start_time=10, end_time=20, cutoff_freq=500)

        This will apply a progressive high-pass filter to the audio segment between 10 and 20 seconds, starting with
        a low cutoff frequency and increasing it progressively to 500 Hz.
        """
        num_fragments = 50
  
        #Init different chuncks of audio
        highpass_audio = self.y[int(start_time*self.sr):int(end_time*self.sr)]
        before_audio = self.y[:int(start_time*self.sr)]
        after_audio = self.y[int(end_time*self.sr):]
        
        #Init progressive filtering with intervals
        cutoff_log = np.log10(cutoff_freq)
        log_intervals = np.logspace(1.7, cutoff_log, num_fragments)
        frame_size = len(highpass_audio) / num_fragments
        
        y_new = []

        #Loop through and apply progressive filter to audio
        for frame,log_interval in zip(range(num_fragments),log_intervals):
            yn = highpass_audio[int(frame*frame_size): int((frame+1)*frame_size)]
            nyquist_freq = 0.5 * self.sr
            normalized_cutoff_freq = log_interval / nyquist_freq
            b, a = scipy.signal.butter(order, normalized_cutoff_freq, btype='high', analog=False)
            yn = scipy.signal.lfilter(b, a, yn)
            yn = self.__fade_between(yn, .003, .003) #Fade between as to prevent pops/clicks
            y_new += (list(yn))

        #Alter after_audio with final cutoff
        normalized_cutoff_freq = log_intervals[-1] / nyquist_freq
        b, a = scipy.signal.butter(order, normalized_cutoff_freq, btype='high', analog=False)
        after_audio = scipy.signal.lfilter(b, a, after_audio)

        #Concat new audio
        y_array = np.array(y_new)
        y_final = np.concatenate([before_audio, y_array, after_audio])

        return Process(y_final)
    
    

    def lowpass_control(self, start_time:int, end_time:int, cutoff_freq:int, order=5) -> object:

        """
        Applies a progressive low-pass filter to an audio segment between specified start and end times.

        Parameters
        ----------
        start_time : int
            The start time in seconds of the segment to apply the low-pass filter.
        end_time : int
            The end time in seconds of the segment to apply the low-pass filter.
        cutoff_freq : int
            The cutoff frequency of the low-pass filter in Hertz. The filter will progressively adjust this frequency.
        order : int, optional
            The order of the Butterworth filter. Default is 5. Higher values result in a steeper roll-off.

        Returns
        -------
        numpy.ndarray or Audio
            If `play` is True, returns an `Audio` object with the processed audio. If `play` is False, returns the raw
            processed audio data as a numpy array.

        Notes
        -----
        The function applies a low-pass filter to the specified audio segment, starting with a high cutoff frequency
        and progressively decreasing it. This is done in 50 intervals, creating a smooth transition effect. The audio
        segments before and after the specified range are not filtered.

        Examples
        --------
        To apply a low-pass filter starting at 10 seconds, ending at 20 seconds, with a cutoff frequency of 500 Hz:

        >>> audio.lowpass_control(start_time=10, end_time=20, cutoff_freq=500)

        This will apply a progressive low-pass filter to the audio segment between 10 and 20 seconds, starting with
        a high cutoff frequency and decreasing it progressively to 500 Hz.
        """
        num_fragments = 50
        
        #Init different chuncks of audio
        highpass_audio = self.y[int(start_time*self.sr):int(end_time*self.sr)]
        before_audio = self.y[:int(start_time*self.sr)]
        after_audio = self.y[int(end_time*self.sr):]

        #Init progressive filtering with intervals
        cutoff_log = np.log10(cutoff_freq)
        log_intervals = np.logspace(4, cutoff_log, num_fragments)
        frame_size = len(highpass_audio) / num_fragments
        
        y_new = [] #New highpass audio list

        #Loop through and apply progressive filter to audio
        for frame,log_interval in zip(range(num_fragments),log_intervals):
            yn = highpass_audio[int(frame*frame_size): int((frame+1)*frame_size)]
            nyquist_freq = 0.5 * self.sr
            normalized_cutoff_freq = log_interval / nyquist_freq
            b, a = scipy.signal.butter(order, normalized_cutoff_freq, btype='low', analog=False)
            yn = scipy.signal.lfilter(b, a, yn)
            yn = self.__fade_between(yn, .003, .003) #Fade between as to prevent pops/clicks
            y_new += (list(yn))
            
        #Alter after_audio with final cutoff
        normalized_cutoff_freq = log_intervals[-1] / nyquist_freq
        b, a = scipy.signal.butter(order, normalized_cutoff_freq, btype='low', analog=False)
        after_audio = scipy.signal.lfilter(b, a, after_audio)

        y_array = np.array(y_new)
        y_final = np.concatenate([before_audio, y_array, after_audio])
        
        return Process(y_final)


# Mix50/mixfifty.py
from .effects import Effects
from .features import Features
from .process import Process
from .transitions import Transitions
import librosa


class MixFifty:
    """
    A class for handling audio processing and effects using various modules.

    Attributes
    ----------
    path1 : str
        The file path for the first audio file.
    path2 : str, optional
        The file path for the second audio file (default is None).
    y1 : ndarray
        The audio time series data for the first audio file.
    sr1 : int
        The sample rate of the first audio file.
    y2 : ndarray, optional
        The audio time series data for the second audio file (default is None).
    sr2 : int, optional
        The sample rate of the second audio file (default is None).

    Parameters
    ----------
    path1 : str
        The file path to the first audio file to be loaded.
    path2 : str, optional
        The file path to the second audio file to be loaded (default is None).
    """
    
    def __init__(self):
        """
        Initializes the MixFifty instance by creating instances of the Effects,
        Process, and Features classes.
        """
        # Initialize all modules
        self.effects = Effects()
        self.features = Features()
        self.transitions = Transitions()

    
    def load_audio(self, path1, path2=None):
        """
        Loads audio data from the specified file paths and passes it to the
        Effects, Features, and Process modules.

        Parameters
        ----------
        path1 : str
            The file path to the first audio file to be loaded.
        path2 : str, optional
            The file path to the second audio file to be loaded (default is None).

        Notes
        -----
        The audio files are loaded using librosa with a sample rate of 22050 Hz.
        If only one audio file is provided, the second audio is set to None.
        """
        self.path1 = path1
        self.path2 = path2
        self.y1, self.sr1 = librosa.load(path1, sr=22050)
        
        if path2:
            self.y2, self.sr2 = librosa.load(path2, sr=22050)
        else:
            self.y2 = None
            self.sr2 = None

        # Pass the loaded audio to classes
        self.effects.set_audio(self.path1, self.y1, self.sr1)
        self.features.set_audio(self.path1, self.y1, self.sr1)
        self.transitions.set_audio(self.path1, self.y1, self.sr1, self.path2, self.y2, self.sr2)


#Work in progress
from Mix50.features import Features
from Mix50.effects import Effects
from Mix50.process import Process
import warnings
import numpy as np
import pandas as pd

warnings.filterwarnings("ignore")

class Transitions:

    def __init__(self):
        
        pass


    def set_audio(self, path1, y1, sr1, path2, y2, sr2):
        '''
        Set the audio from our mixfifty module
        '''
        self.y1 = y1
        self.sr1 = sr1
        self.path1 = path1
        self.y2 = y2
        self.sr2 = sr2
        self.path2 = path2
        
        #We need both features sets and effects within our transitions
        #There is likely a better way to do this
        self.features_audio_1 = Features()
        self.features_audio_2 = Features()
        self.features_audio_1.set_audio(self.path1, self.y1,self.sr1)
        self.features_audio_2.set_audio(self.path2, self.y2,self.sr2)

        self.effects_audio_1 = Effects()
        self.effects_audio_2 = Effects()
        self.effects_audio_1.set_audio(self.path1, self.y1,self.sr1)
        self.effects_audio_2.set_audio(self.path2, self.y2,self.sr2)


    def crossfade(self, 
                  cue_num1:int, 
                  cue_num2:int,
                  filter_type:str, 
                  fade_duration:int=15, 
                  cutoff_frequency:int=0) -> object:
        """
    Apply speed control, filtering, and crossfade effects to two audio tracks based on their beat grids.

    This method processes two audio tracks by aligning their beats, adjusting their speeds to match BPMs, 
    applying specified filters, and performing a crossfade transition between them. The resulting audio 
    is a seamless mix of both tracks with the desired effects applied.

    Parameters:
        cue_num1 (int):
            The cue number for the first audio track. 
            Determines the start time for applying speed control and other effects.
        
        cue_num2 (int):
            The cue number for the second audio track. 
            Determines the start time for the fade in of the second track
        
        filter_type (str):
            The type of filter to apply to the second audio track. 
            Must be one of the following:
                - 'highpass': Apply a high-pass filter.
                - 'lowpass': Apply a low-pass filter.
                - 'none': Do not apply any filter.
        
        fade_duration (int, optional):
            The duration of the transition effect in seconds. 
            Defaults to 15 seconds.
        
        cutoff_frequency (int, optional):
            The cutoff frequency for the filter in Hertz (Hz). 
            Defaults to 0 Hz.

    Returns:
        Process:
            An instance of the `Process` class containing the final mixed audio 
            after applying speed control, filtering, and crossfade effects.

    Raises:
        IndexError:
            If `cue_num1` or `cue_num2` is out of range of the available loop cues 
            in the beat grids of the audio tracks.
        
        ValueError:
            If an invalid `filter_type` is provided. 
            Valid options are 'highpass', 'lowpass', or 'none'.

        """
        
        # Retrieve audio features and data
        beats1, beats2 = self.features_audio_1.beats(), self.features_audio_2.beats()
        bpm1, bpm2 = self.features_audio_1.bpm(), self.features_audio_2.bpm()
        y1, y2 = self.y1, self.y2
        sr1, sr2 = self.sr1, self.sr2

        # Define constants
        fade_duration2 = fade_duration
        sample_rate = 22050
        cutoff_frequency = cutoff_frequency
        
        # Determine start and end times for the crossfade. Use the cue number and raise if out of range
        try:
            loop_cue_values2 = beats2[beats2.loop_cues.notna()].loop_cues.values.tolist()
            loop_cue_values1 = beats1[beats1.loop_cues.notna()].loop_cues.values.tolist()    
            start_time2 = loop_cue_values2[cue_num2]
            end_time2 = loop_cue_values2[cue_num2 + 2]
            start_time1 = loop_cue_values1[cue_num1]
        except IndexError:
            raise IndexError("The cue number is out of range. Please pick a smaller cue number.")

        
        # Apply speed control, high-pass filter, and fade out effects
        audio2 = self.effects_audio_2.speed_control(start_time2, end_time2, bpm2, bpm1).raw_audio()
        self.effects_audio_2.set_audio('./', audio2, sample_rate)

        #Set filter types
        if filter_type == 'highpass':
            audio2 = self.effects_audio_2.highpass_control(start_time2, end_time2, cutoff_frequency).raw_audio()
            self.effects_audio_2.set_audio('./', audio2, sample_rate)
        elif filter_type == 'lowpass':
            audio2 = self.effects_audio_2.lowpass_control(start_time2, end_time2, cutoff_frequency).raw_audio()
            self.effects_audio_2.set_audio('./', audio2, sample_rate)
        elif filter_type == 'none':
            pass
        
        else:
            raise ValueError("Invalid filter type: please choose either 'highpass' or 'lowpass'")

        audio2 = self.effects_audio_2.fade_out(end_time2, fade_duration2).raw_audio() #Start fading at the end of the FX

        #Now that we have our affected audio for song2, we need to calculate a new beatgrid because of altered speed
        #Let's first return the beatgrid starting from our start_transition_time
        affected_beatgrid_2 = beats2[(beats2.beats >= start_time2) & (beats2.beats <= end_time2)]
        
        #Get relevent beats to slice our DF
        last_beat = affected_beatgrid_2.beats.values[-1]
        first_beat = affected_beatgrid_2.beats.values[0]
        time_between_first_last = last_beat - first_beat
        last_beat_affected = time_between_first_last * (bpm2/bpm1) + first_beat
        num_beats = len(affected_beatgrid_2)
        new_beats = np.linspace(first_beat, last_beat_affected, num=num_beats)
        
        # Replace old beats with new beats
        affected_beatgrid_2['beats'] = new_beats
        
        # Update downbeat, loop_cues, and transitions based on new beats
        # The conditions remain the same as they are just updates based on new timing
        affected_beatgrid_2['downbeat'] = np.where(affected_beatgrid_2['downbeat'].notna(), affected_beatgrid_2['beats'], affected_beatgrid_2['downbeat'])
        affected_beatgrid_2['loop_cues'] = np.where(affected_beatgrid_2['loop_cues'].notna(), affected_beatgrid_2['beats'], affected_beatgrid_2['loop_cues'])
        affected_beatgrid_2['transitions'] = np.where(affected_beatgrid_2['transitions'].notna(), affected_beatgrid_2['beats'], affected_beatgrid_2['transitions'])
        
        #Lets get the before-effects beatgrid
        before_affected_beatgrid_2 = beats2[(beats2.beats < start_time2)]
        after_affected_beatgrid_2 = beats2[(beats2.beats > end_time2)]
        
        #Now we have to affect the beatgrid after the transition
        first_beat = after_affected_beatgrid_2.beats.values[0]
        second_beat = after_affected_beatgrid_2.beats.values[1]
        time_between_first_second = second_beat - first_beat
        speed_factor = time_between_first_second * (bpm2/bpm1) #New time between beats accounting for speed change
        num_beats = len(after_affected_beatgrid_2)
        new_beats = [first_beat + (speed_factor*frame) for frame in range(num_beats)]
        
        # Replace old beats with new beats
        after_affected_beatgrid_2['beats'] = new_beats
        
        # Update downbeat, loop_cues, and transitions based on new beats
        # The conditions remain the same as they are just updates based on new timing
        after_affected_beatgrid_2['downbeat'] = np.where(after_affected_beatgrid_2['downbeat'].notna(), after_affected_beatgrid_2['beats'], after_affected_beatgrid_2['downbeat'])
        after_affected_beatgrid_2['loop_cues'] = np.where(after_affected_beatgrid_2['loop_cues'].notna(), after_affected_beatgrid_2['beats'], after_affected_beatgrid_2['loop_cues'])
        after_affected_beatgrid_2['transitions'] = np.where(after_affected_beatgrid_2['transitions'].notna(), after_affected_beatgrid_2['beats'], after_affected_beatgrid_2['transitions'])
        
        #Concat final beatgrid with effects taken place
        final_affected_beatgrid = pd.concat([before_affected_beatgrid_2, affected_beatgrid_2, after_affected_beatgrid_2])

        #Now we have the affected 2nd song, we need to overlay the first.
        new_start_time2 = final_affected_beatgrid[final_affected_beatgrid.loop_cues.notna()].loop_cues.values.tolist()[cue_num2]
        new_end_time2 = final_affected_beatgrid[final_affected_beatgrid.loop_cues.notna()].loop_cues.values.tolist()[cue_num2]
        
        #Break the audio into pieces
        audio2_overlay_len = int((new_end_time2+fade_duration2)*sample_rate)
        audio2_end_sample = audio2_overlay_len+((fade_duration2+5)*sample_rate)
        audio2_beginning_overlay_sample = int((new_end_time2+fade_duration2)*sample_rate)
        audio2_beginning = audio2[:int((new_end_time2+fade_duration2)*sample_rate)]
        audio2_overlay = audio2[audio2_beginning_overlay_sample:audio2_end_sample]
        
        #Get start time of transition song and overlay with other song
        audio1_sample_start = int(start_time1*sample_rate)
        audio1_overlay = y1[audio1_sample_start:audio1_sample_start+((fade_duration2+5)*sample_rate)]
        audio1_overlay = audio1_overlay * np.linspace(0,1,len(audio1_overlay)) #Fade the second song in. Without this it is abrupt
        audio1_end = y1[audio1_sample_start+((fade_duration2+5)*sample_rate):]
        
        #Full overlay 
        full_overlay = audio2_overlay + audio1_overlay
        full_audio = list(audio2_beginning) + list(full_overlay) + list(audio1_end)

        return Process(full_audio)


from setuptools import setup, find_packages

with open('README.md', encoding='utf8') as f:
    long_description = f.read()

setup(
    name='Mix50',
    version='0.5',
    description='Fast and simple DJ and audio effects in Python with Librosa',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.11',
        'Topic :: Multimedia :: Sound/Audio :: Analysis'
    ],
    url='https://github.com/smullins998/Mix50',
    author='Sean Mullins',
    author_email='smullins998@gmail.com',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'librosa>=0.10.2.post1',
        'numpy<2',
        'pandas',
        'SoundFile',
        'pyrubberband',
        'scipy',
        'pydub',
        'scikit-learn',
        'ipython',
        'essentia',
        'matplotlib',
        'sounddevice',
    ],
    extras_require={
        'ffmpeg': ['ffmpeg-python'],
        'portaudio': ['pyaudio'],
    },
    keywords=['python', 'audio', 'DSP', 'DJ', 'audio analysis'],
    include_package_data=True,
    zip_safe=False
)


import soundfile as sf
from IPython.display import Audio
import numpy as np

class Process:
    '''
    This class supports the effects module as a wrapper for further manipulations.
    '''
    
    def __init__(self, audio):
        
        '''
        Set the audio from our mixfifty module
        '''
           
        self.y, self.sr = audio, 22050
       
        
    def play(self):
        '''
        Plays the audio using IPython's Audio display.

        '''
        return Audio(self.y, rate=22050)
    
    def save(self, file_path: str):
        '''
        Saves the audio data to a file in PCM 24-bit format. MP3 is not supported! Please use .WAV!
        '''
        sf.write(file_path, self.y, 22050, 'PCM_24')
        return f"Exported to {file_path}"
    
    def raw_audio(self) -> np.ndarray:
        '''
        Return the raw audio as an nd.array
        '''
        return self.y



import librosa
import numpy as np
import pandas as pd
import soundfile as sf
import pyrubberband as pyrb
from pydub import AudioSegment
from sklearn.cluster import KMeans
from IPython.display import Audio
import scipy
from essentia.standard import *
import essentia.streaming as ess
import essentia.standard as es
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
import sounddevice as sd
from typing import Optional, List
from Mix50.process import Process

    
class Features:
    
    def __init__(self):

        pass
    
    
    def set_audio(self, path, y1, sr1):
        '''
        Set the audio from our mixfifty module
        '''
        self.y = y1
        self.sr = sr1
        self.path = path
  
    
    def beats(self) -> pd.DataFrame:
        """
        Identify transition points in an audio sample, similar to DJ's beat matching.

        Analyzes the audio to find points where transitions are likely to occur based on changes in energy or beat. 
        Uses tempo and beat detection, filtering, RMS feature extraction, and clustering to identify these points.

        Returns
        -------
        pd.DataFrame
            DataFrame of transition points in seconds. Includes columns for beats, downbeats, loop cues, and transitions.
        """

        # Detect tempo and beats using librosa
        tempo, beats = librosa.beat.beat_track(y=self.y, sr=self.sr)
        beat_times = [round(i, 3) for i in librosa.frames_to_time(beats)]

        # Low-pass filter setup
        cutoff_hz = 70
        nyquist = 0.5 * self.sr
        normal_cutoff = cutoff_hz / nyquist
        b, a = scipy.signal.butter(8, normal_cutoff, btype='low')

        # Apply low-pass filter to audio
        y_low = scipy.signal.lfilter(b, a, self.y)
        y_perc = librosa.effects.percussive(self.y, margin=5)
        y_perc = scipy.signal.lfilter(b, a, y_perc)

        # RMS and smoothing parameters
        window_duration = 0.05
        frame_length = int(window_duration * self.sr)
        hop_length = frame_length // 2

        # RMS calculation and smoothing
        rms_df = pd.DataFrame({
            'RMS': librosa.feature.rms(y=self.y, frame_length=frame_length, hop_length=hop_length)[0],
            'SMA': pd.Series(librosa.feature.rms(y=self.y, frame_length=frame_length, hop_length=hop_length)[0]).rolling(150, center=True).mean().fillna(0),
            'low': librosa.feature.rms(y=y_low, frame_length=frame_length, hop_length=hop_length)[0],
            'SMA_low': pd.Series(librosa.feature.rms(y=y_low, frame_length=frame_length, hop_length=hop_length)[0]).rolling(150, center=True).mean().fillna(0),
            'perc': librosa.feature.rms(y=y_perc, frame_length=frame_length, hop_length=hop_length)[0],
            'SMA_perc': pd.Series(librosa.feature.rms(y=y_perc, frame_length=frame_length, hop_length=hop_length)[0]).rolling(40, center=True).mean().fillna(0),
        })

        # Calculate percentage change and completion
        rms_df['PCT'] = rms_df['RMS'].pct_change()
        rms_df['completion'] = rms_df.index / len(rms_df)

        # Scale and weight features
        scaler = StandardScaler()
        features = ['SMA_low', 'SMA_perc', 'SMA', 'low']
        X_scaled = scaler.fit_transform(rms_df[features])
        X_weighted = X_scaled * [1, 0.4, 1, 0.2]

        # Perform KMeans clustering
        kmeans = KMeans(n_clusters=2, random_state=42)
        rms_df['clusters'] = kmeans.fit_predict(X_weighted)

        # Identify transition points
        rms_df['cue'] = rms_df['clusters'].diff().ne(0).astype(int)
        rms_df['seconds'] = round(rms_df['completion'] * librosa.get_duration(y=self.y), 2)

        # Retrieve cues and find closest values to beats
        onset = rms_df[rms_df.cue == 1]['seconds'].values
        if len(onset) == 0:
            return pd.DataFrame(columns=['beats', 'downbeat', 'loop_cues', 'transitions'])

        differences = [abs(value - onset[0]) for value in beat_times]
        idx_pos = differences.index(min(differences))
        closest_value = beats[idx_pos]

        # Calculate downbeats, loop cues, and transitions
        downbeat = beat_times[idx_pos::-4][::-1][:-1] + beat_times[idx_pos::4]
        loop_cues = beat_times[idx_pos::-16][::-1][:-1] + beat_times[idx_pos::16]
        transitions = [min(loop_cues, key=lambda x: abs(x - value)) for value in onset]

        # Create beat DataFrame with results
        beat_df = pd.DataFrame({'beats': beat_times})
        beat_df['downbeat'] = beat_df['beats'].apply(lambda x: x if x in downbeat else None)
        beat_df['loop_cues'] = beat_df['beats'].apply(lambda x: x if x in loop_cues else None)
        beat_df['transitions'] = beat_df['beats'].apply(lambda x: x if x in transitions else None)
        self.beat_df = beat_df
        
        return beat_df
    
    
    def transition_points(self) -> np.ndarray:
        '''
        Retrieve transition points from the processed beat DataFrame.

        This method extracts transition points from the `beat_df` DataFrame, which should be populated by the `.beats()` method.
        Transition points represent significant changes or cues in the audio data that were identified during the beat analysis.

        Returns
        -------
        np.ndarray
            An array of transition points, which are significant moments in the audio where transitions occur. 
            NaN values are excluded from the output.

        Raises
        ------
        AttributeError
            If the `beat_df` attribute is not set (i.e., if `.beats()` has not been called before this method), an error message is printed
            and the method returns an empty list.

        Notes
        -----
        - Ensure that the `.beats()` method is executed before calling this method to populate the `beat_df` attribute.
        - The `beat_df` DataFrame is expected to have a column named 'transitions' which contains the transition points.

        Examples
        --------
        To retrieve transition points after running the `.beats()` method:

        >>> audio.beats()  # First, compute beats and transitions
        >>> transitions = audio.transition_points()  # Retrieve transition points
        >>> print(transitions)
        [12.34, 56.78, ...]

        '''

        try:
            # Check if data exists
            beat_df = self.beat_df
        except AttributeError:
            # Print custom error message and return early to prevent plotting
            print("Error: Please run the method '.beats()' first.")
            return
        
        transition_points = beat_df.transitions.values
        transition_points = transition_points[~np.isnan(transition_points)]
        
        return list(transition_points)
            
    
    def show_transition_points(self):
        '''
        Show transition points on a waveplot
        '''

        try:
            # Check if data exists
            beat_df = self.beat_df
        except AttributeError:
            # Print custom error message and return early to prevent plotting
            print("Error: Please run the method '.beats()' first.")
            return

        # If no exception, proceed with plotting
        plt.figure(figsize=(14, 5))
        librosa.display.waveshow(self.y, sr=self.sr, alpha=0.7)

        for t in list(beat_df.transitions):
            plt.axvline(x=t, color='r', linestyle='--', label='Transition Point')

        plt.title('Waveplot with Transition Points')
        plt.xlabel('Time (s)')
        plt.ylabel('Amplitude')
        plt.legend(['Transition Point'], loc='upper right')
        plt.show()
        
        
    def beat_matching(self) -> pd.DataFrame:
        '''
        Retrieve downbeats using MonoLoader 
        '''
        
        #Load Audio
        audio = MonoLoader(filename=self.path)()

        # Compute beat positions and BPM
        rhythm_extractor = RhythmExtractor2013(method="multifeature")
        bpm, beats, beats_confidence, _, beats_intervals = rhythm_extractor(audio)

        beat_df = pd.DataFrame({'beats':beats})

        return beat_df


    def bpm(self) -> int:
        '''
        Retrieve BPM using MonoLoader 
        '''

        #Load audio
        audio = es.MonoLoader(filename=self.path)()

        # Compute beat positions and BPM.
        rhythm_extractor = es.RhythmExtractor2013(method="multifeature")
        bpm, beats, beats_confidence, _, beats_intervals = rhythm_extractor(audio)
        
        return round(bpm,0)
    

    def key(self) -> str:
        '''
        Retrieve Key using Essentia 
        '''

        # Initialize Monoloader
        loader = ess.MonoLoader(filename=self.path)
        framecutter = ess.FrameCutter(frameSize=4096, hopSize=2048, silentFrames='noise')
        windowing = ess.Windowing(type='blackmanharris62')
        spectrum = ess.Spectrum()
        spectralpeaks = ess.SpectralPeaks(orderBy='magnitude',
                                          magnitudeThreshold=0.00001,
                                          minFrequency=20,
                                          maxFrequency=3500,
                                          maxPeaks=60)

        hpcp = ess.HPCP()
        hpcp_key = ess.HPCP(size=36, # We will need higher resolution for Key estimation.
                            referenceFrequency=440, # Assume tuning frequency is 44100.
                            bandPreset=False,
                            minFrequency=20,
                            maxFrequency=1500,
                            weightType='cosine',
                            nonLinear=False,
                            windowSize=1.)

        key = ess.Key(profileType='edma', # Use profile for electronic music.
                      numHarmonics=4,
                      pcpSize=36,
                      slope=0.6,
                      usePolyphony=True,
                      useThreeChords=True)

        # Use pool to store data.
        pool = essentia.Pool()

        # Connect streaming algorithms.
        loader.audio >> framecutter.signal
        framecutter.frame >> windowing.frame >> spectrum.frame
        spectrum.spectrum >> spectralpeaks.spectrum
        spectralpeaks.magnitudes >> hpcp.magnitudes
        spectralpeaks.frequencies >> hpcp.frequencies
        spectralpeaks.magnitudes >> hpcp_key.magnitudes
        spectralpeaks.frequencies >> hpcp_key.frequencies
        hpcp_key.hpcp >> key.pcp
        hpcp.hpcp >> (pool, 'tonal.hpcp')
        key.key >> (pool, 'tonal.key_key')
        key.scale >> (pool, 'tonal.key_scale')
        key.strength >> (pool, 'tonal.key_strength')

        essentia.run(loader)

        return "Estimated key and scale:", pool['tonal.key_key'] + " " + pool['tonal.key_scale']